/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2005, 2012  All Rights Reserved.
 */

package com.ibm.lang.management;

/**
 * The IBM-specific interface for monitoring the operating system where the
 * virtual machine is running.
 * 
 * @author gharley
 * @since 1.5
 */
public interface OperatingSystemMXBean extends
        java.lang.management.OperatingSystemMXBean {

    /**
     * Returns the total available physical memory on the system in bytes.
     * 
     * @return the total available physical memory on the system in bytes.
     */
    public long getTotalPhysicalMemory();

    /**
     * Returns the collective capacity of the virtual processors in
     * the partition the VM is running in. The value returned is in
     * units of 1% of a physical processor's capacity, so a value of
     * 100 is equal to 1 physical processor. In environments without
     * such partitioning support, this call will return
     * getAvailableProcessors() * 100.
     * 
     * @return the collective capacity of the virtual processors available
     * to the VM
     */
    public int getProcessingCapacity();

    /**
     * Returns total amount of time the process has been scheduled or
     * executed so far in both kernel and user modes. Returns -1 if the
     * value is unavailable on this platform or in the case of an error.
     *
     * @return process cpu time in 100 ns units
     */
    public long getProcessCpuTime();
    
    /**
     * Returns total amount of time the process has been scheduled or executed 
     * so far in both kernel and user modes (in nano seconds). Returns -1 if the
     * value is unavailable on this platform or in the case of
     * an error.
     *
     * @return process cpu time in ns units
     */
    public long getProcessCpuTimeByNS();

    /**
     * <p>Returns the CPU utilization: fraction of the time spent in user or system/privileged
     * mode since the last call to getSystemCpuLoad(). Other values include:</p>
     * <ul>
     *     <li>ERROR_VALUE (-1): Insufficient time (less than 10 ms) has elapsed since the first
     * call to getSystemCpuLoad(). Alternatively, an operating system error such as clock rollover,
     * has occurred.</li> 
     *     <li>INSUFFICIENT_PRIVILEGE (-2). This operation is not supported due to insufficient
     * user privileges.</li>
     *     <li>UNSUPPORTED_VALUE (-3): The operation is not supported on this platform.</li>
     *     <li>An internal problem has occurred.</li>
     * </ul>
     * <p>getSystemCpuLoad might not return the same value that is reported by operating system
     * utilities such as Unix "top" or Windows task manager.</p>
     *  
     * @return value between 0 and 1.0, or a negative value in the case of an error.
     */

    public double getSystemCpuLoad();

    /**
     * Returns the amount of free physical memory at current instance on the
     * system in bytes. Returns -1 if the value is unavailable on this
     * platform or in the case of an error.
     *
     * @return amount of physical memory available in bytes
     */
    public long getFreePhysicalMemorySize();

    /**
     * Returns the amount of virtual memory used by the process in bytes,
     * including physical memory and swap space. Returns -1 if the value
     * is unavailable on this platform or in the case of an error.
     *
     * @return amount of virtual memory used by the process in bytes
     */
    public long getProcessVirtualMemorySize();

    /**
     * Returns the amount of private memory used by the process in bytes.
     * Returns -1 if the value is unavailable on this platform or in the
     * case of an error.
     *
     * @return amount of private memory used by the process in bytes
     */
    public long getProcessPrivateMemorySize();

    /**
     * Returns the amount of physical memory being used by the process
     * in bytes. Returns -1 if the value is unavailable on this platform
     * or in the case of an error.
     *
     * @return amount of physical memory being used by the process in bytes
     */
    public long getProcessPhysicalMemorySize();

    /**
     * Returns the total amount of swap space in bytes.
     *
     * @return the total amount of swap space in bytes.
     */
    public long getTotalSwapSpaceSize();

    /**
     * Returns the amount of free swap space in bytes.
     *
     * @return the amount of free swap space in bytes.
     */
    public long getFreeSwapSpaceSize();

    /**
     * Returns the "recent cpu usage" for the Java Virtual Machine process.
     * This value is a double in the [0.0,1.0] interval. A value of 0.0 means
     * that none of the CPUs were running threads from the JVM process during
     * the recent period of time observed, while a value of 1.0 means that all
     * CPUs were actively running threads from the JVM 100% of the time
     * during the recent period being observed. Threads from the JVM include
     * the application threads as well as the JVM internal threads. All values
     * between 0.0 and 1.0 are possible depending on the activities within the
     * JVM process and the system as a whole. First call to the method always 
     * returns -1.0 marking the starting point. If the Java Virtual Machine
     * recent CPU usage is not available, the method returns a negative value.
     *
     * @return the "recent cpu usage" for the Java Virtual Machine process;
     * -1.0 for first call; a negative value if not available.
     * @since   1.7
     */
    public double getProcessCpuLoad();

}

/*
 * $Log$
 * Revision 1.5  2005/05/12 00:34:50  pschurch
 * add getProcessingCapacity
 *
 * Revision 1.4  2005/04/18 10:39:42  gharley
 * Removal of all attributes bar PhysicalMemory from the proprietary interface. Agreed with Paul Church and Tim Preece.
 *
 * Revision 1.3  2005/02/24 16:47:35  gharley
 * Incorrect method name. Replaced getCommittedPhysicalMemory with the
 * method called getCommittedVirtualMemory
 *
 * Revision 1.2  2005/02/14 13:21:05  gharley
 * D'oh ! The extension interface ought to extend the standard interface.
 * Revision 1.1 2005/02/14 12:10:18 gharley
 * Initial implementation.
 * 
 */
