/*
 * Copyright (c) 1995, 2011, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

/**
 * =======================================================================
 * Module Information:
 *
 * DESCRIPTION: This class is used to convert sockets from the TCP to
 * a type that is supported by the underlying network provider before
 * accept or connect happens.
 * Much of this code is derived from sun.net.sdp.SdpProvider class.
 * 
 * This package private final class provides RDMA_CAPABILITY.
 * =======================================================================
 */

package java.net;

import java.io.File;
import java.io.IOException;
import java.security.AccessController;
import java.util.Formatter;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Scanner;
import java.util.Enumeration;
import java.lang.reflect.Method;

import sun.security.action.GetPropertyAction;

final class NetworkSelector {
	// indicates if the RDMA network provider is enabled
	private static final boolean enabled;
	private static final String osName;
	private static final String osArch;
	// rules when the RDMA protocol is used
	private static LinkedList<Rule> bindRules;
	private static LinkedList<Rule> connectRules;
	private static LinkedList<Rule> acceptRules;
	// network provider to name mapping
	private static HashMap<String, NetworkProvider> providerMap;
        
	// Avoid class instantiation
	private NetworkSelector() {
		// do nothing
	}
	
	/**
     * Helper class to parse the port spec:
     *   <port-spec> = ("*" | port)[ "-" ("*" | port) ]
     */
	private static class PortRange {
    	private final int portStart;
    	private final int portEnd;
    	private static final int MAX_PORT = 65535;
    	
    	public PortRange(String s) throws NumberFormatException {
    		// check for range indicator
    		int pos = s.indexOf('-');
    		if (pos < 0) {
    			// only starting port field?
    			boolean all = s.equals("*");
    			portStart = all ? 0 : Integer.parseInt(s);
    			portEnd = all ? MAX_PORT : portStart;
    		} else {
    			// both start and end port fields?
    			String low = s.substring(0,pos);
    			if (low.length() == 0) {
    				low = "*";
    			}
    			String high = s.substring(pos+1);
    			if (high.length() == 0) {
    				high = "*";
    			}
    			portStart = low.equals("*") ? 0 : Integer.parseInt(low);
    			portEnd = high.equals("*") ? MAX_PORT : Integer.parseInt(high);
    		}
    	}
    	
    	// check whether the specified port falls within the identified range
    	public boolean match(int port) {
    		return (port >= this.portStart && port <= this.portEnd);
    	}
    }
		
	/**
     * Helper class to parse the host spec:
     *   <host-spec> = (hostname | ipaddress ["/" prefix])
     */
    private static class Host {
    	private final boolean anyHost;
    	private final byte[] addressAsBytes;
    	private final int prefixByteCount;
    	private final byte mask;
    	private final boolean isLoopBack;
    	private static List<String> localAddresses = new LinkedList<String>();

    	final static Host[] allHosts = new Host[] {new Host() {
    		public boolean match(InetAddress address) {
    			return true;
    		}
    	}};
    	
    	// retrieve all the network interfaces on this host
    	static {
            try {
                Enumeration<NetworkInterface> nis = NetworkInterface.getNetworkInterfaces();

                // walk through the network interfaces to obtain IP addresses			
                while (nis.hasMoreElements()) {
                    NetworkInterface ni = nis.nextElement();
                    Enumeration<InetAddress> inetAddrs = ni.getInetAddresses();
                    while (inetAddrs.hasMoreElements()) {
                        localAddresses.add(inetAddrs.nextElement().getHostAddress());
                    }
                }
            } catch (SocketException e) {
                // Ignore...
            }
        }
    	
    	// constructor for any host
    	private Host() {
    		this.anyHost = true;
    		this.addressAsBytes = null;
    		this.prefixByteCount = 0;
    		this.mask = 0;
    		this.isLoopBack = false;
    	}
    	
    	// check whether the IP address is loopback or local
    	static public boolean isLoopBackorLocal(InetAddress addr) {
    		for (String address: localAddresses) {
    			if (address.equals(addr.getHostAddress()) || addr.isLoopbackAddress()) {
    				return true;
    			}
    		}
    		return false;
    	}
       	
    	// constructor for a host with specific IP address and prefix
    	private Host(InetAddress address, int prefix) {
    		this.anyHost = false;
    		this.addressAsBytes = address.getAddress();
    		this.prefixByteCount = prefix >> 3;
    		this.mask = (byte)(0xff << (8 - (prefix % 8)));
    		this.isLoopBack = isLoopBackorLocal(address);
    	}
    	
    	// check whether the specified IP address belongs to this host
    	boolean match(InetAddress address) {
    		/* null address is not comparable with rule to take care of multihomed host
    		 * where only specific ips are RDMA enabled. Fall back to TCP
    		 */
    		if (address == null) {
    			return false;
    		}
    		if (anyHost) {
    			return true;
    		}
    		// retrieve the raw IP address
    		byte[] candidate = address.getAddress();
    		// check address length
    		if (candidate.length != addressAsBytes.length) {
    			return false;
    		}
    		// check address bytes
    		for (int i = 0; i < prefixByteCount; i++) {
    			if (candidate[i] != addressAsBytes[i]) {
    				return false;
    			}
    		}
    		// check remaining bits
    		if ((prefixByteCount < addressAsBytes.length) &&
    				((candidate[prefixByteCount] & mask) != (addressAsBytes[prefixByteCount] & mask))) {
    			return false;
    		}
    		return true;
    	}
    	
    	// parse the specified host spec and return an arry of hosts
    	static Host[] parse(String hostString) {
    		if (hostString.equals("*")) {
    			return allHosts;
    		}
    		
    		LinkedList<Host> hosts = new LinkedList<Host>();
    		int pos = hostString.indexOf('/');
    		try {
    			if (pos < 0) {
    				// hostname or ipaddress (no prefix)
    				// get all the ip addresses for the specified host
    				InetAddress[] addresses = InetAddress.getAllByName(hostString);
    				for (InetAddress address: addresses) {
    					int prefix = (address instanceof Inet4Address) ? 32 : 128;
    					hosts.addLast(new Host(address, prefix));
    				}
    			} else {
    				// ipaddress/prefix
    				// get the ip address for the specified host
    				InetAddress address = InetAddress.getByName(hostString.substring(0, pos));
    				int prefix = -1;
    				try {
    					prefix = Integer.parseInt(hostString.substring(pos + 1));
    					if (address instanceof Inet4Address) {
    						// must be 1-31
    						if (prefix < 0 || prefix > 32) prefix = -1;
    					} else {
    						// must be 1-128
    						if (prefix < 0 || prefix > 128) prefix = -1;
    					}
    				} catch (NumberFormatException e) {
    					// Ignore...
    				}
    				if (prefix > 0) {
    					hosts.addLast(new Host(address, prefix));
    				} else {
    					fail("Malformed prefix '%s'", hostString);
    					return null;
    				}
    			}
    		} catch (UnknownHostException uhe) {
    			fail("Unknown host or malformed IP address '%s'", hostString);
    			return null;
    		}
    		return hosts.toArray(new Host[hosts.size()]);
    	}
    	
    	// getter method for loopback flag
    	public boolean getIsLoopBack() {
    		return this.isLoopBack;
    	}
    }
	
    /**
     * A rule for matching an accept or connect request
     */
    private static interface Rule {
    	boolean match(InetAddress localAddress, int localPort, InetAddress remoteAddress);
    	// getter methods for retrieving various rule related attributes
    	SocketAction getAction();
    	PortRange getPortRange();
    	Host getPrimaryHost();
    	Host getSecondaryHost();
    	NetworkProvider getNetworkProvider();
    }
	
    /**
     * Helper class to match host-port rule
     */
    private static class HostPortRule implements Rule {
    	private SocketAction action;
    	private PortRange range;
    	private Host host;
    	private Host remote;
    	private NetworkProvider provider;
    	
    	// constructor
    	HostPortRule(SocketAction action, Host host, PortRange range, Host remote, NetworkProvider provider) {
    		this.action = action;
    		this.host = host;
    		this.range = range;
    		this.remote = remote;
    		this.provider = provider;
    	}
    	
    	public boolean match(InetAddress localAddress, int localPort, InetAddress remoteAddr) {
    		if (!host.match(localAddress)) {
    			return false;
    		}

    		if (!range.match(localPort)) {
    			return false;
    		}
    	
    		// remoteAddr will be null in cases where the rule is for a bind/connect
    		if (remoteAddr == null || remote == null) {
    			return true;
    		}
    		return remote.match(remoteAddr);
    	}
    	
    	public SocketAction getAction() {
    		return action;
    	}
    	
    	public PortRange getPortRange() {
    		return range;
    	}
    	
    	public Host getPrimaryHost() {
    		return host;
    	}
    	
    	public Host getSecondaryHost() {
    		return remote;
    	}
    	
    	public NetworkProvider getNetworkProvider() {
    		return provider;
    	} 	
    }
    
    /**
     * Add rule to the appropriate list
     */
    private static void addRule(Rule rule) {
    	switch(rule.getAction()) {
    	case BIND:
    		bindRules.addLast(rule);
    		break;
    	case ACCEPT:
    		acceptRules.addLast(rule);
    		// for each accept rule also add a bind rule
    		bindRules.addLast(new HostPortRule(SocketAction.BIND, rule.getPrimaryHost(), rule.getPortRange(),
    				rule.getSecondaryHost(), rule.getNetworkProvider()));
    		break;
    	case CONNECT:
    		connectRules.addLast(rule);
    		break;
    	default:
    		// do nothing
    		break;
    	}
    }
      
    /**
     * Helper method to throw runtime exception in failing to parse
     * a rule field.
     */
    private static void fail(String msg, Object... args) {
    	Formatter f = new Formatter();
    	f.format(msg, args);
    	throw new RuntimeException(f.out().toString());
    }
    
    /**
     * Load rules from the specified configuration file.
     * 
     * Each non-blank or non-comment line must have the following format:
     *   <entry>         = <net-spec> <sp> <connect-entry> | <accept-entry>
     *   <net-spec>      = "rdma"
     *   <sp>            = 1*LWSP-char
     *   <connect-entry> = "connect" <host-spec> <sp> <port-spec>
     *   <host-spec>     = (hostname | ipaddress ["/" prefix])
     *   <port-spec>     = ("*" | port)[ "-" ("*" | port) ]
     *   <accept-entry>  = "accept" <host-spec> <sp> <port-spec> <sp> <client-spec> [<sp> 1*<client-spec>]
     *   <client-spec>   = (clientname | ipaddress ["/" prefix])
     */
    private static void loadRulesFromFile(String file) throws IOException {
    	Scanner scanner = new Scanner(new File(file));
    	try {
    		while (scanner.hasNextLine()) {
    			String line = scanner.nextLine().trim();
    			
    			// skip blank lines and comments
    			if (line.length() == 0 || line.charAt(0) == '#') {
    				continue;
    			}
    			
    			// must have at least 4 fields
    			String[] s = line.split("\\s+");
    			if (s.length < 4) {
    				fail("Malformed line '%s'", line);
    				continue;
    			}
    			
    			// first field is the network identifier
    			NetworkProvider provider = loadProvider(s[0]);
    			if (provider == null) {
    				fail("Network provider for '%s' could not be loaded", s[0]);
    				continue;
    			}
    			
    			// second field is the action ("connect" or "accept")
    			SocketAction action = SocketAction.parse(s[1]);
    			if (action == null) {
    				fail("SocketAction '%s' not recognized", s[1]);
    				continue;
    			}
    			
    			if (action == SocketAction.ACCEPT && s.length < 5) {
    				fail("Less than expected number of arguments '%s'", line);
    				continue;
    			}
    			
    			// third field is the host spec
    			Host[] primaryHosts = Host.parse(s[2]);
    			if (primaryHosts == null) {
    				continue;
    			}
    			
    			// fourth field is the port spec
    			PortRange range = null;
    			try {
    				range = new PortRange(s[3]);
    			} catch (NumberFormatException nfe) {
    				fail("Malformed port range '%s'", s[3]);
    				continue;
    			}
    			   			
    			/* fifth and above fields (if exist) contain client addresses
    			 * create one rule for each host-client combination
    			 */
    			if (s.length > 4) {
    				for (int i = 4; i < s.length; i++) {
    					Host[] h = Host.parse(s[i]);
    					if (h != null) {
    						for (int p = 0; p < primaryHosts.length; p++) {
    							for (int j = 0; j < h.length; j++) {
    								if (h[j].getIsLoopBack()) {
    									continue;
    								}
    								addRule(new HostPortRule(action, primaryHosts[p], range, h[j], provider));
    							}
    						}
    					}
    				}
    			} else {
    				// client address portion is null
    				for (int p = 0; p < primaryHosts.length; p++) {
    					if (primaryHosts[p].getIsLoopBack()) {
    						continue;
    					}
    					addRule(new HostPortRule(action, primaryHosts[p], range, null, provider));
    				}
    			}
    		}
    	} finally {
    		scanner.close();
    	}
    }
    
    /**
	 * Initialize the protocol usage rules
	 */
	static boolean initializeRules() {
		// flag indicating whether rules are loaded from the configuration file
		boolean rulesLoaded = false;
		// retrieve the configuration file
		String confFile = AccessController.doPrivileged(new GetPropertyAction("java.net.conf"));
		if (confFile != null) {
			bindRules = new LinkedList<Rule>();
			connectRules = new LinkedList<Rule>();
			acceptRules = new LinkedList<Rule>();
			providerMap = new HashMap<String, NetworkProvider>();
			try {
				loadRulesFromFile(confFile);
				rulesLoaded = true;
			} catch (IOException e) {
				// Ignore...
			}
			if (rulesLoaded) {
				// Register a shutdown hook to cleanup the network resources
				Runtime.getRuntime().addShutdownHook(new Thread() {
					public void run() {
						Iterator<NetworkProvider> it = providerMap.values().iterator();
						while (it.hasNext()) {
							NetworkProvider networkProvider = it.next();
							networkProvider.cleanup();
						}
					}
				});
			} else {
				bindRules = null;
				connectRules = null;
				acceptRules = null;
				providerMap = null;
			}
		}
		return rulesLoaded;
	}
	
	// this implementation currently supported on linux-x86_64 platform only
	static {
		// flag indicating whether rules are initialized
		boolean rulesInited = false;
		GetPropertyAction propAct = new GetPropertyAction("os.name");
		osName = AccessController.doPrivileged(propAct);
		propAct = new GetPropertyAction("os.arch");
		osArch = AccessController.doPrivileged(propAct);
		if (osName.equalsIgnoreCase("linux") && osArch.equalsIgnoreCase("amd64")) {
			rulesInited = initializeRules();
		}
		enabled = rulesInited;
	}
	
	/**
	 * Find the first rule that matches the specified attributes in a rule set
	 */
	private static Rule findFirstMatchingRule(List<Rule> rules, InetAddress primaryHost,
			int port, InetAddress secondaryHost) {
		Iterator<Rule> it = rules.iterator();
		while (it.hasNext()) {
			Rule rule = it.next();
			if (rule.match(primaryHost, port, secondaryHost)) {
				return rule;
			}
		}
		return null;
	}
	    
	/**
	 * Retrieve the underlying network provider
	 */
	public static NetworkProvider getNetworkProvider(SocketAction action,
			InetAddress primaryHost, int port, InetAddress secondaryHost) {
		Rule matchedRule = null;
		if (!enabled) {
			return null;
		}
		switch (action) {
		case BIND:
			matchedRule = findFirstMatchingRule(bindRules, primaryHost, port, secondaryHost);
			break;
		case ACCEPT:
			matchedRule = findFirstMatchingRule(acceptRules, primaryHost, port, secondaryHost);
			break;
		case CONNECT:
			matchedRule = findFirstMatchingRule(connectRules, primaryHost, port, secondaryHost);
			break;
		default:
			break;
		}
		if (matchedRule != null) {
			return matchedRule.getNetworkProvider();
		}
		return null;
	}
	
    /**
     * Load the specified network provider - currently,
     * we only check RDMA network provider
     */
    private static NetworkProvider loadProvider(String name) {
        NetworkProvider provider = null;
        if (providerMap.containsKey(name)) {
            provider = providerMap.get(name);
    	} else if (name.equalsIgnoreCase("rdma")) {
            try {
            	// Let's rely on the bootstrap class loader to locate and load the RDMANetworkProvider class
            	Class netProvider = Class.forName("java.net.RDMANetworkProvider", true, null);
                provider = (NetworkProvider)netProvider.newInstance();
                Method intializeProvider = netProvider.getMethod("initialize", (Class[])null);
                intializeProvider.invoke(provider, (Object[])null);
            } catch (Exception e) {
                return null;
            }
            providerMap.put(name, provider);
    	}
    	return provider;
    }
}
