/*
 * Copyright (c) 1995, 2011, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

/**
 * =======================================================================
 * Module Information:
 *
 * DESCRIPTION: This class implements RDMA socket by extending from
 * the PlainSockImpl.
 * 
 * This package private class provides RDMA_CAPABILITY.
 * =======================================================================
 */

package java.net;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileDescriptor;

class RDMASocketImpl extends PlainSocketImpl {
	// associated input/output streams and abstract plain socket impl
	private RDMASocketInputStream socketInputStream = null;
	private RDMASocketOutputStream socketOutputStream = null;
	private AbstractPlainSocketImpl abstractPlainSocketImpl = null;
	// store the pointer of the native data structure related to this socket
	private long socketDataPtr;
	private final SocketImpl oldImpl;
	private boolean created = false;
		
	// native methods
	static native void initProtoRDMA();
	native long socketCreateRDMA(FileDescriptor fdObj) throws IOException;
	native void socketConnectRDMA(InetAddress address, int port, long socketDataPtr) throws IOException;
	native void socketBindRDMA(InetAddress address, int port, long socketDataPtr) throws IOException;
	native void socketListenRDMA(int count, long socketDataPtr) throws IOException;
	native long socketPostAccept(SocketImpl acceptedSocketImpl, long serverSocketDataPtr) throws IOException;
	native void socketCloseRDMA(boolean useDeferredClose, long socketDataPtr) throws IOException;
	native void socketShutdownRDMA(int howto, long socketDataPtr) throws IOException;
	native void socketSendUrgentDataRDMA(int data, long socketDataPtr) throws IOException;
	
	// do native initialization only once
	static {
		initProtoRDMA();
	}
	
	// helper method to create RDMA socket and set data pointer field
	private void socketCreatePrivate(FileDescriptor fdObj) throws IOException {
		socketDataPtr = socketCreateRDMA(fdObj);
		created = true;
	}
	
	/**
	 * Constructs a RDMASocketImpl instance
	 */
	RDMASocketImpl(SocketImpl oldImpl) {
		this.oldImpl = oldImpl;
	}
	
	/**
	 * Creates an RDMA stream socket
	 */
	@Override
	protected void create(boolean stream) throws IOException {
		if (!stream) {
			throw new UnsupportedOperationException("Must be a stream socket for RDMA!");
		}
		super.create(stream);
	}
	
	/**
	 * Store abstract plain socket impl that is used
	 */
	void setAbstractPlainSocketImpl(AbstractPlainSocketImpl impl) {
		abstractPlainSocketImpl = impl;
	}
	
	/**
	 * Return true if this RDMA socket is already closed or close is pending
	 */
	@Override
	public boolean isClosedOrPending() {
		if (super.isClosedOrPending()) {
			return true;
		}
		if (socketDataPtr == 0) {
			return true;
		}
		return false;
	}
	
	/**
	 * Get input stream for this RDMA socket.
	 */
	@Override
	protected synchronized InputStream getInputStream() throws IOException {
		if (isClosedOrPending()) {
			throw new IOException("Socket Closed");
		}
		if (inputShut()) {
			throw new IOException("Socket input is shutdown");
		}
		if (socketInputStream == null) {
			socketInputStream = new RDMASocketInputStream(this);
		}
		return socketInputStream;
	}
	
	/**
	 * Get output stream for the RDMA socket.
	 */
	@Override
	protected synchronized OutputStream getOutputStream() throws IOException {
		if (isClosedOrPending()) {
			throw new IOException("Socket Closed");
		}
		if (outputShut()) {
			throw new IOException("Socket output is shutdown");
		}
		if (socketOutputStream == null) {
			socketOutputStream = new RDMASocketOutputStream(this);
		}
		return socketOutputStream;
	}
	
	/**
	 * z/OS specific function hence null implementation to satisfy base class
	 */
	void processNRException(InetAddress address, int port, int backlogQ) {
		// do nothing
	}
	
	/**
	 * Create an RDMA stream socket
	 */
	@Override
	void socketCreate(boolean isServer) throws IOException {
		super.socketCreate(isServer);
		socketCreatePrivate(null);
	}
	
	/**
	 * Connect this RDMA socket to the specified port on the specified host
	 */
	@Override
	void socketConnect(InetAddress address, int port, int timeout) throws IOException {
		/* The socket connect can happen lately after creating a plain unconnected socket.
		 * In such a scenario, we need to ensure that RDMA portion of the socket is also
         * created and linked to the plain socket before attempting the connect operation.
		 */
		if (!created && (oldImpl != null)) {
			socketCreatePrivate(oldImpl.getFileDescriptor());
		}
		super.socketConnect(address, port, timeout);
		try {
			socketConnectRDMA(address, port, socketDataPtr);
		} catch (Exception e) {
			socketShutdownRDMA(0, socketDataPtr);
			socketDataPtr = 0;
			throw new IOException("RDMA Connect Failed!");
		}
	}
	
	/**
	 * Bind this RDMA socket to the specified IP address and port number
	 */
	@Override
	void socketBind(InetAddress address, int port) throws IOException {
		/* The socket bind can happen lately after a plain server socket creation.
		 * If that is the case, then we need to ensure that RDMA portion of the server
		 * socket is also created and linked to the plain server socket before attempting
		 * the bind operation.
		 */
		if (!created && (oldImpl != null)) {
			socketCreatePrivate(oldImpl.getFileDescriptor());
			created = true;
		}
		super.socketBind(address, port);
		socketBindRDMA(address, port, socketDataPtr);
	}
	
	/**
	 * Listen for incoming connection indications on this RDMA socket
	 * with the maximum queue length of 'count'
	 */
	@Override
	void socketListen(int count) throws IOException {
		super.socketListen(count);
		socketListenRDMA(count, socketDataPtr);
	}
	
	/**
	 * Close this RDMA socket
	 */
	@Override
	void socketClose0(boolean useDeferredClose) throws IOException {
		// close the RDMA socket followed by the plain socket
		socketCloseRDMA(useDeferredClose, socketDataPtr);
		super.socketClose0(useDeferredClose);
	}
	
	/**
	 * Shutdown this RDMA connection
	 */
	@Override
	void socketShutdown(int howto) throws IOException {
		// shutdown the RDMA socket followed by the plain socket
		socketShutdownRDMA(howto, socketDataPtr);
		super.socketShutdown(howto);
	}
	
	/**
	 * Send urgent data on this RDMA socket connection
	 */
	@Override
	void socketSendUrgentData(int data) throws IOException {
		socketSendUrgentDataRDMA(data, socketDataPtr);
	}

	/**
	 * Returns the local port to which this socket is bound.
	 */
	@Override
	protected int getLocalPort() {
		if (abstractPlainSocketImpl != null) {
			return(abstractPlainSocketImpl.getLocalPort());
		} else {
			return(super.getLocalPort());
		}
	}
	
	/**
	 * Returns the remote port to which this socket is connected.
	 */
	@Override
	protected int getPort() {
		if (abstractPlainSocketImpl != null) {
			return(abstractPlainSocketImpl.getPort());
		} else {
			return(super.getPort());
		}
	}
	
	/**
	 * Getter method for the RDMA socket pointer
	 */
	long getSocketDataPtr() {
		return socketDataPtr;
	}
	
	/**
	 * Setter method for the RDMA socket pointer
	 */
	void setSocketDataPtr(long socketDataPtr) {
		this.socketDataPtr = socketDataPtr;
	}
}
