/*
 * Copyright (c) 1997, 2011, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

/**
 * =======================================================================
 * Module Information:
 *
 * DESCRIPTION: This class implements RDMA socket input stream by
 * extending SocketInputStream.  Note that this class should <b>NOT<b>
 * be made public.
 * 
 * This package private class provides RDMA_CAPABILITY.
 * =======================================================================
 */

package java.net;

import java.io.FileDescriptor;
import java.io.IOException;
import sun.net.ConnectionResetException;

class RDMASocketInputStream extends SocketInputStream {
	// native methods
	private native static void init();
	private native int socketRead0(FileDescriptor fd, long sockDataPtr, byte b[], int off,
			int len, int timeout) throws IOException;
	
	// end of file flag
	private boolean eof;
	// RDMA socket impl associated with this input stream
	private RDMASocketImpl impl = null;
	
	// perform class load-time initialization only once
	static {
		init();
	}
	
	/**
	 * Creates a new RDMA socket input stream.
	 */
	RDMASocketInputStream(RDMASocketImpl impl) throws IOException {
		super(impl);
		this.impl = impl;
	}
	
	/**
	 * Reads into a byte array the specified number of bytes at
	 * the specified offset.
	 */
	int read(byte b[], int off, int length, int timeout) throws IOException {
		// read byte count
		int n;
		// reset flag
		boolean gotReset = false;
		
		// eof reached?
		if (eof) {
			return -1;
		}
		
		// connection reset
		if (impl.isConnectionReset()) {
			throw new SocketException("Connection reset");
		}
		
		// bounds check
		if (length <=0 || off < 0 || off + length > b.length) {
			if (length == 0) {
				return 0;
			}
			throw new ArrayIndexOutOfBoundsException();
		}
		
		// acquire file descriptor and do the read
		FileDescriptor fd = impl.acquireFD();
		try {
			n = socketRead0(fd, impl.getSocketDataPtr(), b, off, length, timeout);
			if (n > 0) {
				return n;
			}
		} catch (ConnectionResetException cre) {
			gotReset = true;
		} finally {
			impl.releaseFD();
		}
		
		/* A connection reset is received but there may be bytes
		 * still left on the buffered socket
		 */
		if (gotReset) {
			impl.setConnectionResetPending();
			impl.acquireFD();
			try {
				n = socketRead0(fd, impl.getSocketDataPtr(), b, off, length, timeout);
				if (n > 0) {
					return n;
				}
			} catch (ConnectionResetException cre) {
				// do nothing
			} finally {
				impl.releaseFD();
			}
		}
		
		// the socket might have been closed
		if (impl.isClosedOrPending()) {
			throw new SocketException("Socket closed");
		}
		
		// the connection could have been reset
		if (impl.isConnectionResetPending()) {
			impl.setConnectionReset();
		}
		if (impl.isConnectionReset()) {
			throw new SocketException("Connection reset");
		}
		
		// set eof flag
		eof = true;
		return -1;
	}
}
