/*
 * Copyright (c) 1997, 2011, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

/**
 * =======================================================================
 * Module Information:
 *
 * DESCRIPTION: This class implements RDMA socket output stream by
 * extending SocketOutputStream.  Note that this class should <b>NOT<b>
 * be made public.
 * 
 * This package private class provides RDMA_CAPABILITY.
 * =======================================================================
 */
package java.net;

import java.io.FileDescriptor;
import java.io.IOException;
import sun.net.ConnectionResetException;

class RDMASocketOutputStream extends SocketOutputStream {
	// native methods
	private native static void init();
	private native void socketWrite0(FileDescriptor fd, long sockDataPtr, byte b[], int off,
			int len) throws IOException;
	
	// RDMA socket impl associated with this output stream
	private RDMASocketImpl impl = null;
	// temporary buffer for one byte write
	private byte temp[] = new byte[1];
	
	// perform class load-time initialization only once
	static {
		init();
	}
	
	/**
	 * Creates a new RDMA socket output stream.
	 */
	RDMASocketOutputStream(RDMASocketImpl impl) throws IOException {
		super(impl);
		this.impl = impl;
	}
	
	/**
	 * Writes to the RDMA socket with appropriate locking of the file descriptor.
	 */
	private void socketWrite(byte b[], int off, int len) throws IOException {
		// bounds checking
		if (len <= 0 || off < 0 || off + len > b.length) {
			if (len == 0) {
				return;
			}
			throw new ArrayIndexOutOfBoundsException();
		}
		FileDescriptor fd = impl.acquireFD();
		try {
			socketWrite0(fd, impl.getSocketDataPtr(), b, off, len);
		} catch (SocketException se) {
			if (se instanceof ConnectionResetException) {
				impl.setConnectionResetPending();
				se = new SocketException("Connection reset");
			}
			if (impl.isClosedOrPending()) {
				throw new SocketException("Socket closed");
			} else {
				throw se;
			}
		} finally {
			impl.releaseFD();
		}
	}
	
	/**
	 * Writes a byte to the RDMA socket.
	 */
	public void write(int b) throws IOException {
		temp[0] = (byte)b;
		socketWrite(temp, 0, 1);
	}
	
	/**
	 * Writes the specified buffer contents to the RDMA socket.
	 */
	public void write(byte b[]) throws IOException {
		socketWrite(b, 0, b.length);
	}
	
	/**
	 * Writes the specified number of bytes from the buffer
	 * at the specified offset.
	 */
	public void write(byte b[], int off, int len) throws IOException {
		socketWrite(b, off, len);
	}
}
